import math

import torch
import torch.nn.functional as F
from torch import nn
from torch_geometric.nn import global_mean_pool

from gvp import GVP, GVPConvLayer, LayerNorm


def transpose_qkv(X, num_heads):
    """Transposition for parallel computation of multiple attention heads.

    Defined in :numref:`sec_multihead-attention`"""
    # Shape of input `X`:
    # (`batch_size`, no. of queries or key-value pairs, `num_hiddens`).
    # Shape of output `X`:
    # (`batch_size`, no. of queries or key-value pairs, `num_heads`,
    # `num_hiddens` / `num_heads`)
    X = X.reshape(X.shape[0], X.shape[1], num_heads, -1)

    # Shape of output `X`:
    # (`batch_size`, `num_heads`, no. of queries or key-value pairs,
    # `num_hiddens` / `num_heads`)
    X = X.permute(0, 2, 1, 3)

    # Shape of `output`:
    # (`batch_size` * `num_heads`, no. of queries or key-value pairs,
    # `num_hiddens` / `num_heads`)
    return X.reshape(-1, X.shape[2], X.shape[3])


def sequence_mask(X, valid_len, value=0):
    """Mask irrelevant entries in sequences.

    Defined in :numref:`sec_seq2seq_decoder`"""
    maxlen = X.size(1)
    mask = torch.arange((maxlen), dtype=torch.float32,
                        device=X.device)[None, :] < valid_len[:, None]
    X[~mask] = value
    return X


def masked_softmax(X, valid_lens):
    """Perform softmax operation by masking elements on the last axis.

    Defined in :numref:`sec_attention-scoring-functions`"""
    # `X`: 3D tensor, `valid_lens`: 1D or 2D tensor
    if valid_lens is None:
        return nn.functional.softmax(X, dim=-1)
    else:
        shape = X.shape
        if valid_lens.dim() == 1:
            valid_lens = torch.repeat_interleave(valid_lens, shape[1])
        else:
            valid_lens = valid_lens.reshape(-1)
        # On the last axis, replace masked elements with a very large negative
        # value, whose exponentiation outputs 0
        X = sequence_mask(X.reshape(-1, shape[-1]), valid_lens,
                          value=-1e6)
        return nn.functional.softmax(X.reshape(shape), dim=-1)


def transpose_output(X, num_heads):
    """Reverse the operation of `transpose_qkv`.

    Defined in :numref:`sec_multihead-attention`"""
    X = X.reshape(-1, num_heads, X.shape[1], X.shape[2])
    X = X.permute(0, 2, 1, 3)
    return X.reshape(X.shape[0], X.shape[1], -1)


def mean_tensor(combined_tensor, lens):
    """
    tensor will be divided according to lens, and then merge with means
    :param combined_tensor:
    :param lens: tensor lens
    :return:
    """
    means = [combined_tensor[i, :l, :].mean(0) for i, l in enumerate(lens)]
    return torch.stack(means, dim=0)


def reshape_tensors(seq, seq_lens, pad_idx):
    """
    Combine tensors of different lengths into one tensor, where the excess is filled with pad_idx
    :param seq:
    :param seq_lens:
    :param pad_idx:
    :return:
    """
    if len(seq_lens) != 1:
        batch_size = len(seq_lens)
        seq_max_len = max(seq_lens)
        total = 0
        shape = list(seq.shape)
        shape.insert(0, batch_size)
        shape[1] = seq_max_len
        seqs = torch.full(shape, pad_idx, dtype=seq.dtype).to(seq.device)
        for i, seq_len in enumerate(seq_lens):
            seqs[i, :seq_len] = seq[total:total + seq_len]
            total += seq_len
    else:
        seqs = seq.unsqueeze(0)
    return seqs


class AddNorm(nn.Module):
    """Layer normalization after residual concatenation"""

    def __init__(self, normalized_shape, dropout, **kwargs):
        super(AddNorm, self).__init__(**kwargs)
        self.dropout = nn.Dropout(dropout)
        self.ln = nn.LayerNorm(normalized_shape)

    def forward(self, X, Y):
        return self.ln(self.dropout(Y) + X)


class PositionWiseFFN(nn.Module):
    """Location-based feedforward networks"""

    def __init__(self, ffn_num_input, ffn_num_hiddens, ffn_num_outputs,
                 **kwargs):
        super(PositionWiseFFN, self).__init__(**kwargs)
        self.dense1 = nn.Linear(ffn_num_input, ffn_num_hiddens)
        self.relu = nn.ReLU()
        self.dense2 = nn.Linear(ffn_num_hiddens, ffn_num_outputs)

    def forward(self, X):
        return self.dense2(self.relu(self.dense1(X)))


class DotProductAttention(nn.Module):
    """Scaled dot product attention.

    Defined in :numref:`subsec_additive-attention`"""

    def __init__(self, dropout, **kwargs):
        super(DotProductAttention, self).__init__(**kwargs)
        self.dropout = nn.Dropout(dropout)

    # Shape of `queries`: (`batch_size`, no. of queries, `d`)
    # Shape of `keys`: (`batch_size`, no. of key-value pairs, `d`)
    # Shape of `values`: (`batch_size`, no. of key-value pairs, value
    # dimension)
    # Shape of `valid_lens`: (`batch_size`,) or (`batch_size`, no. of queries)
    def forward(self, queries, keys, values, valid_lens=None):
        d = queries.shape[-1]
        # Set `transpose_b=True` to swap the last two dimensions of `keys`
        scores = torch.bmm(queries, keys.transpose(1, 2)) / math.sqrt(d)
        self.attention_weights = masked_softmax(scores, valid_lens)
        return torch.bmm(self.dropout(self.attention_weights), values)


class MultiHeadAttention(nn.Module):
    """Multi-head attention.

    Defined in :numref:`sec_multihead-attention`"""

    def __init__(self, key_size, query_size, value_size, num_hiddens,
                 num_heads, dropout, bias=False, **kwargs):
        super(MultiHeadAttention, self).__init__(**kwargs)
        self.num_heads = num_heads
        self.attention = DotProductAttention(dropout)
        self.W_q = nn.Linear(query_size, num_hiddens, bias=bias)
        self.W_k = nn.Linear(key_size, num_hiddens, bias=bias)
        self.W_v = nn.Linear(value_size, num_hiddens, bias=bias)
        self.W_o = nn.Linear(num_hiddens, num_hiddens, bias=bias)

    def forward(self, queries, keys, values, valid_lens):
        # Shape of `queries`, `keys`, or `values`:
        # (`batch_size`, no. of queries or key-value pairs, `num_hiddens`)
        # Shape of `valid_lens`:
        # (`batch_size`,) or (`batch_size`, no. of queries)
        # After transposing, shape of output `queries`, `keys`, or `values`:
        # (`batch_size` * `num_heads`, no. of queries or key-value pairs,
        # `num_hiddens` / `num_heads`)
        queries = transpose_qkv(self.W_q(queries), self.num_heads)
        keys = transpose_qkv(self.W_k(keys), self.num_heads)
        values = transpose_qkv(self.W_v(values), self.num_heads)

        if valid_lens is not None:
            # On axis 0, copy the first item (scalar or vector) for
            # `num_heads` times, then copy the next item, and so on
            valid_lens = torch.repeat_interleave(
                valid_lens, repeats=self.num_heads, dim=0)

        # Shape of `output`: (`batch_size` * `num_heads`, no. of queries,
        # `num_hiddens` / `num_heads`)
        output = self.attention(queries, keys, values, valid_lens)

        # Shape of `output_concat`:
        # (`batch_size`, no. of queries, `num_hiddens`)
        output_concat = transpose_output(output, self.num_heads)
        return self.W_o(output_concat)


class FunctionPredictor(torch.nn.Module):
    def __init__(self, input_size, output_size, drop_rate=0.1):
        super(FunctionPredictor, self).__init__()

        hidden_size = 4 * output_size
        self.input_layer = nn.Linear(input_size, hidden_size)
        self.output_layer = nn.Linear(hidden_size, output_size)
        self.relu = nn.ReLU()
        self.dropout = nn.Dropout(p=drop_rate)

    def forward(self, x):
        x = self.input_layer(x)
        x = self.relu(x)
        x = self.dropout(x)
        x = self.output_layer(x)
        return x


class EncoderBlock(nn.Module):
    """transformer Encoder Block

    Defined in :numref:`sec_transformer`"""

    def __init__(self, key_size, query_size, value_size, num_hiddens,
                 norm_shape, ffn_num_input, ffn_num_hiddens, num_heads,
                 dropout, use_bias=False, **kwargs):
        super(EncoderBlock, self).__init__()
        self.attention = MultiHeadAttention(
            key_size, query_size, value_size, num_hiddens, num_heads, dropout,
            use_bias)
        self.addnorm1 = AddNorm(norm_shape, dropout)
        self.ffn = PositionWiseFFN(
            ffn_num_input, ffn_num_hiddens, num_hiddens)
        self.addnorm2 = AddNorm(norm_shape, dropout)

    def forward(self, X, valid_lens):
        Y = self.addnorm1(X, self.attention(X, X, X, valid_lens))
        return self.addnorm2(Y, self.ffn(Y))


class GVPNet(nn.Module):
    def __init__(self, node_in_dim, node_h_dim,
                 edge_in_dim, edge_h_dim, hidden_dim, num_gvp_layers, gvp_drop_rate, seq_in=False):
        super(GVPNet, self).__init__()

        # gvp
        # node_in_dim = (node_in_dim[0] + esm_dim, node_in_dim[1])
        if seq_in:
            self.W_s = nn.Embedding(20, 20)
            node_in_dim = (node_in_dim[0] + 20, node_in_dim[1])

        self.W_v = nn.Sequential(
            GVP(node_in_dim, node_h_dim, activations=(None, None)),
            LayerNorm(node_h_dim)
        )
        self.W_e = nn.Sequential(
            GVP(edge_in_dim, edge_h_dim, activations=(None, None)),
            LayerNorm(edge_h_dim)
        )

        self.gvp_layers = nn.ModuleList(
            GVPConvLayer(node_h_dim, edge_h_dim, drop_rate=gvp_drop_rate)
            for _ in range(num_gvp_layers))

        self.W_out = GVP(node_h_dim, (hidden_dim, 0), activations=(None, None))

    def forward(self, h_V, h_E, edge_index, seq=None):
        if seq is not None:
            seq = self.W_s(seq)
            h_V = (torch.cat([h_V[0], seq], dim=-1), h_V[1])
        h_V = self.W_v(h_V)
        h_E = self.W_e(h_E)
        for layer in self.gvp_layers:
            h_V = layer(h_V, edge_index, h_E)
        return self.W_out(h_V)


class PredGONet(nn.Module):
    """
    ESM+GVP+PPN Net 不使用ESM作为氨基酸结点特征 ESM 与PPN的特征维度相同 使用AddAndNorm进行sequence embedding 的相加
    """

    def __init__(self, num_class, aa_node_in_dim,
                 aa_node_h_dim,
                 aa_edge_in_dim, aa_edge_h_dim, gvp_out_dim, num_gvp_layers, ppn_num_heads,
                 num_ppn_layers, drop_rate=0.2):
        super(PredGONet, self).__init__()
        self.gvp_net = GVPNet(aa_node_in_dim, aa_node_h_dim,
                              aa_edge_in_dim, aa_edge_h_dim, gvp_out_dim, num_gvp_layers, drop_rate)
        self.esm_dim = 1280
        self.ppn_mlp = nn.ModuleList()
        self.addnorm = AddNorm(normalized_shape=[self.esm_dim], dropout=drop_rate)

        self.ppn_net = nn.ModuleList([EncoderBlock(key_size=self.esm_dim, query_size=self.esm_dim,
                                                   value_size=self.esm_dim, num_hiddens=self.esm_dim,
                                                   norm_shape=[self.esm_dim], ffn_num_input=self.esm_dim,
                                                   ffn_num_hiddens=4 * self.esm_dim, num_heads=ppn_num_heads,
                                                   dropout=drop_rate) for _ in range(num_ppn_layers)])
        self.output_layer = FunctionPredictor(self.esm_dim + gvp_out_dim, num_class,
                                              drop_rate=drop_rate)

    @staticmethod
    def batch_tag_tensor(num_tensor):
        r = []
        for i, v in enumerate(num_tensor):
            r.extend([i] * v)
        return torch.tensor(r, device=num_tensor.device)

    def forward(self, sequence_embedding, aa_h_V, aa_h_E, aa_edge_index, num_aa, ppi_tensor, num_ppi):
        aa_tag_tensor = self.batch_tag_tensor(num_aa)
        ppi_tensor = reshape_tensors(ppi_tensor, num_ppi, 0)
        sequence_embedding = sequence_embedding.view(sequence_embedding.shape[0] // self.esm_dim, self.esm_dim)
        for layer in self.ppn_net:
            ppi_tensor = layer(ppi_tensor, num_ppi)
        ppi_tensor = ppi_tensor[:, 0, :]
        # aa_h_V = (torch.cat([aa_h_V[0], sequence_embedding], dim=1), aa_h_V[1])
        # sequence_embedding = global_mean_pool(sequence_embedding, aa_tag_tensor)
        aa_h_V = self.gvp_net(aa_h_V, aa_h_E, aa_edge_index)
        aa_h_V = global_mean_pool(aa_h_V, aa_tag_tensor)
        sequence_embedding = self.addnorm(sequence_embedding, ppi_tensor)
        output = self.output_layer(torch.cat([sequence_embedding, aa_h_V], dim=1))
        return output
